#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include "patoh.h"



#define NCONST        1


/*
#define _DEBUG1
*/

#define umax(a, b)	(((a) >= (b)) ? (a) : (b))

#define usRandom(s)     srand(s)
#define uRandom(l)      (rand() % l)


int *ReadPartVec(char *fname, int *resvec, int n, int k);


void usage(char *name)
{
 printf("usage: %s <hypergraph-filename> <#parts> [#runs]\n", name);
 exit(1);
}


void PrintInfo(char *pname, int _k, int *partweights, int cut, int _nconst)
{
 double             *avg, *maxi;
 int                i, j;

 printf("\n-------------------------------------------------------------------");
 printf("\n Partitioner: %s", pname);
 
 printf("\n %d-way cutsize = %d \n", _k, cut);
 
#ifdef _DEBUG1
 printf("\nPartWeights are:\n");
#endif
 avg = (double *) malloc(sizeof(double)*_nconst);
 maxi = (double *) malloc(sizeof(double)*_nconst);
 for (i=0; i<_nconst; ++i)
     maxi[i] = avg[i] = 0.0;     
 for (i=0; i<_k; ++i)
       for (j=0; j<_nconst; ++j)
         avg[j] += partweights[i*_nconst+j];
  for (i=0; i<_nconst; ++i)
     {
     maxi[i] = 0.0;     
     avg[i] /= (double) _k;
     }
 
 for (i=0; i<_k; ++i)
     {
#ifdef _DEBUG1
     printf("\n %3d :", i);
#endif
     for (j=0; j<_nconst; ++j)
         {
         double im= (double)((double)partweights[i*_nconst+j] - avg[j]) / avg[j];
         
         maxi[j] = umax(maxi[j], im);
#ifdef _DEBUG1
         printf("%10d ", partweights[i*_nconst+j]);
#endif
         }     
     }
#ifdef _DEBUG1
 printf("\n Average Weights are:");
 printf("\n      ");
 for (i=0; i<_nconst; ++i)
     printf("%10.1lf ", avg[i]);
#endif
 printf("\n MaxImbals are (as %%):");
 printf("\n Imbal");
 for (i=0; i<_nconst; ++i)
     printf("%10.1lf ", 100.0*maxi[i]);
 
 free(maxi);
 free(avg);
}


#define MAXTRY 5

int main(int argc, char *argv[])
{
 PaToH_Parameters args;
 int             _c, _n, _nconst, *cwghts, *nwghts, 
     *xpins, *pins, *partvec, cut, *partweights,
     i;
 char           fname[512];
 int            run, maxrun=1;
 double         avgcost[MAXTRY];
 int mt1=9, mt2=11;
 

 for (i=0; i<MAXTRY; ++i)
     avgcost[i] = 0.0;

 if (argc<3)
     usage(argv[0]);
 if (argc>3)
     maxrun = atoi(argv[3]);
 if (argc>4)
     mt1 = atoi(argv[4]);
 if (argc>5)
     mt2 = atoi(argv[5]);
 

 usRandom(time(NULL));
 

 sprintf(fname, "%s.fix", argv[1]);
 PaToH_Read_Hypergraph(argv[1], &_c, &_n, &_nconst, &cwghts, 
                       &nwghts, &xpins, &pins);

 printf("===============================================================================\n");
 
 printf("Hypergraph %10s -- #Cells=%6d  #Nets=%6d  #Pins=%8d #Const=%2d\n",
        argv[1], _c, _n, xpins[_n], _nconst);
 
 PaToH_Initialize_Parameters(&args, PATOH_CONPART, 
                             PATOH_SUGPARAM_DEFAULT);


 args._k = atoi(argv[2]);
 partvec =  (int *) malloc(_c*sizeof(int));
 partweights  = (int *) malloc(args._k*sizeof(int));

     /*
 
 if (PaToH_Check_Hypergraph(_c, _n, 1, cwghts, nwghts, xpins, pins))
     return -1;
     */
      

 PaToH_Alloc(&args, _c, _n, NCONST, cwghts, nwghts, xpins, pins);

 

 for (run=0; run<maxrun; ++run)
     {
     args.crs_alg=mt1;
     ReadPartVec(fname, partvec, _c, args._k);

     PaToH_Partition_with_FixCells(&args, _c, _n /* try -(10-run) */, cwghts, nwghts,  
                                   xpins, pins, partvec, partweights, &cut);
         /* PrintInfo("With FixCells MT=9 ", args._k, partweights, cut, 1); */
     printf("\n MT=%d: %10d ", mt1, cut);
     
     avgcost[0] += cut;
     for (i=0; i<args._k; ++i)
         if (partvec[i]!=i)
             {
             int j;
             
             printf("\n result partvec:");
             for (j=0; j<args._k; ++j)
                 printf("\nPartvec[%d]=%d", j, partvec[j]);
             break;
             }
     
     args.crs_alg=mt2;
     ReadPartVec(fname, partvec, _c, args._k);

     PaToH_Partition_with_FixCells(&args, _c, _n /* try -(10-run) */, cwghts, nwghts,  
                                   xpins, pins, partvec, partweights, &cut);
/*     PrintInfo("With FixCells MT=11 ", args._k, partweights, cut, 1); */
     printf(" MT=%d: %10d ", mt2, cut);
     avgcost[1] += cut;
     for (i=0; i<args._k; ++i)
         if (partvec[i]!=i)
             {
             int j;
             
             printf("\n result partvec:");
             for (j=0; j<args._k; ++j)
                 printf("\nPartvec[%d]=%d", j, partvec[j]);
             break;
             }    

     }
 for (i=0; i<MAXTRY; ++i)
     avgcost[i] /= maxrun;
 printf("\n %3d run(s) avg. cost MT=%d: %10.2lf   MT=%d: %10.2lf", maxrun, mt1, avgcost[0], mt2, avgcost[1]);
 

free(partweights);
free(partvec);
  
PaToH_Free();
printf("\n");
 
return 0;
}



#define MAXLINE 32768


int *ReadPartVec(char *fname, int *resvec, int n, int k)
{
    int	i;
    
    if (access(fname, 0)) {
        for (i=0; i<k; ++i)
            resvec[i] = i;
        for (; i<n; ++i)
            resvec[i] = -1;             
    } else {
        FILE	*fp;
        char 	*line, *newstr, *oldstr;
        
        
        fp = fopen(fname, "r");

        line = (char*) malloc(MAXLINE);
        for (i=0; i<n;)
        {
            int w;

            do			/* Skip Comments */
            {
                fgets(line, MAXLINE, fp);
            } while (line[0] == '%');
            oldstr = line;
            newstr = NULL;

            if (strlen(line) >= MAXLINE-5) {
                fprintf(stderr, "read_vector : Buffer for fgets not big enough.");
                exit(1);
            }

            w = strtol(oldstr, &newstr, 10);
            while (oldstr!=newstr)
            {
                resvec[i++] = w;
                oldstr = newstr;
                w = strtol(oldstr, &newstr, 10);
            }
        }
        free(line);
        fclose(fp);
    }
return resvec;
}

